import pandas as pd
from neuprint import Client

from .. import Motif
from .Neo4jExecutor import Neo4jExecutor

_LOOKUP = {
    "INHIBITS": "ConnectsTo",
    "EXCITES": "ConnectsTo",
    "SYNAPSES": "ConnectsTo",
    "INH": "ConnectsTo",
    "EXC": "ConnectsTo",
    "SYN": "ConnectsTo",
    "DEFAULT": "ConnectsTo",
}

_DEFAULT_ENTITY_LABELS = {
    "node": "Neuron",
    "edge": _LOOKUP,
}


class NeuPrintExecutor(Neo4jExecutor):
    """
    A NeuPrintExecutor may be used to access an existing neuPrint server.

    This class converts a DotMotif motif object into a neuPrint-compatible
    query. Not all neuPrint datatypes or query types are available, but this
    adds complete support for DotMotif motif searches by passing raw Cypher
    queries to the neuPrint server over the HTTP API.

    Note that the neuPrint default timeout is quite short, and slower motif
    queries may not run in time.

    """

    def __init__(self, host: str, dataset: str, token: str) -> None:
        """
        Create a new NeuPrintExecutor that points to a deployed neuPrint DB.

        Arguments:
            host (str): The host of the neuPrint server (for example,
                'neuprint.janelia.org')
            dataset (str): The name of the dataset to reference (for example,
                'hemibrain:v1.1`)
            token (str): The user's neuPrint access token. To retrieve this
                token, go to https://[host]/account.

        Returns:
            None

        """
        self._created_container = False
        self.host = host
        self.dataset = dataset
        self.token = token
        self.client = Client(host, dataset=self.dataset, token=self.token)

    def run(self, cypher: str) -> pd.DataFrame:
        """
        Run an arbitrary cypher command.

        You should usually ignore this, and use .find() instead.

        Arguments:
            cypher (str): The command to run

        Returns:
            The result of the cypher query

        """
        return self.client.fetch_custom(cypher)

    def count(self, motif: Motif, limit=None) -> int:
        """
        Count a motif in a larger graph.

        Arguments:
            motif (dotmotif.Motif): The motif to search for

        Returns:
            int: The count of this motif in the host graph

        """
        qry = self.motif_to_cypher(
            motif, count_only=True, static_entity_labels=_DEFAULT_ENTITY_LABELS
        )
        if limit:
            qry += f" LIMIT {limit}"
        res = self.client.fetch_custom(qry)
        print(res)
        return int(res.to_numpy())

    def find(self, motif: Motif, limit=None) -> pd.DataFrame:
        """
        Find a motif in a larger graph.

        Arguments:
            motif (dotmotif.Motif): The motif to search for

        Returns:
            pd.DataFrame: The results of the search

        """
        qry = self.motif_to_cypher(motif, static_entity_labels=_DEFAULT_ENTITY_LABELS)
        if limit:
            qry += f" LIMIT {limit}"
        return self.client.fetch_custom(qry)

    @staticmethod
    def motif_to_cypher(
        motif: Motif, count_only: bool = False, static_entity_labels: dict = None
    ) -> str:
        """
        Convert a motif to neuprint-flavored Cypher.

        This is currently a thin passthrough for Neo4jExecutor.motif_to_cypher.

        """
        static_entity_labels = static_entity_labels or _DEFAULT_ENTITY_LABELS
        return Neo4jExecutor.motif_to_cypher(motif, count_only, static_entity_labels)




#!/usr/bin/env python3

from typing import List
from lark import Lark, Transformer
import networkx as nx
import os

from ...utils import untype_string
from .. import Parser
from ...validators import Validator


dm_parser = Lark(open(os.path.join(os.path.dirname(__file__), "grammar.lark"), "r"))


class DotMotifTransformer(Transformer):
    """
    This transformer converts a parsed Lark tree into a networkx.MultiGraph.
    """

    def __init__(self, validators: List[Validator] = None, *args, **kwargs) -> None:
        self.validators = validators if validators else []
        self.macros: dict = {}
        self.G = nx.MultiDiGraph()
        self.edge_constraints: dict = {}
        self.node_constraints: dict = {}
        self.dynamic_edge_constraints: dict = {}
        self.dynamic_node_constraints: dict = {}
        self.automorphisms: list = []
        super().__init__(*args, **kwargs)

    def transform(self, tree):
        self._transform_tree(tree)
        return (
            self.G,
            self.edge_constraints,
            self.node_constraints,
            self.dynamic_edge_constraints,
            self.dynamic_node_constraints,
            self.automorphisms,
        )

    def edge_clauses(self, tup):
        attrs = {}
        for key, op, val in tup:
            if key not in attrs:
                attrs[key] = {}
            if op not in attrs[key]:
                attrs[key][op] = []
            attrs[key][op].append(val)
        return attrs

    def edge_clause(self, tup):
        key, op, val = tup
        val = untype_string(val)
        return str(key), str(op), val

    def node_constraint(self, tup):

        if len(tup) == 4:
            # This is of the form "Node.Key [OP] Value"
            node_id, key, op, val = tup
            node_id = str(node_id)
            key = str(key)
            op = str(op)
            val = untype_string(val)
            if node_id not in self.node_constraints:
                self.node_constraints[node_id] = {}
            if key not in self.node_constraints[node_id]:
                self.node_constraints[node_id][key] = {}
            if op not in self.node_constraints[node_id][key]:
                self.node_constraints[node_id][key][op] = []
            self.node_constraints[node_id][key][op].append(val)

        elif len(tup) == 5:
            # This is of the form "ThisNode.Key [OP] ThatNode.Key"
            this_node_id, this_key, op, that_node_id, that_key = tup

            this_node_id = str(this_node_id)
            this_key = str(this_key)
            that_node_id = str(that_node_id)
            that_key = str(that_key)
            op = str(op)

            if this_node_id not in self.dynamic_node_constraints:
                self.dynamic_node_constraints[this_node_id] = {}

            if this_key not in self.dynamic_node_constraints[this_node_id]:
                self.dynamic_node_constraints[this_node_id][this_key] = {}
            if op not in self.dynamic_node_constraints[this_node_id][this_key]:
                self.dynamic_node_constraints[this_node_id][this_key][op] = []
            self.dynamic_node_constraints[this_node_id][this_key][op].append(
                (that_node_id, that_key)
            )

        else:
            raise ValueError("Something is wrong with the node comparison", tup)

    def macro_node_constraint(self, tup):

        if len(tup) == 4:
            node_id, key, op, val = tup
            node_id = str(node_id)
            key = str(key)
            op = str(op)
            val = untype_string(val)
            return ("node_constraint", node_id, key, op, val)

        if len(tup) == 5:
            this_node, this_key, op, that_node, that_key = tup
            this_node = str(this_node)
            this_key = str(this_key)
            that_node = str(that_node)
            that_key = str(that_key)
            op = str(op)
            return ("node_constraint", this_node, this_key, op, that_node, that_key)

        raise ValueError(
            "Something is wrong in this macro with the node comparison ", tup
        )

    def key(self, key):
        return str(key)

    def op(self, operator):
        return {
            "=": "==",
            "==": "==",
            ">=": ">=",
            "<=": "<=",
            "<": "<",
            ">": ">",
            "<>": "!=",
            "!=": "!=",
        }[operator]

    def edge(self, tup):
        if len(tup) == 3:
            u, rel, v = tup
            attrs = {}
        elif len(tup) == 4:
            u, rel, v, attrs = tup
        u = str(u)
        v = str(v)
        for val in self.validators:
            val.validate(self.G, u, v, rel["type"], rel["exists"])
        if self.G.has_edge(u, v):
            # There are existing edges. Only add a new one if it's unique
            # from all existing ones.
            candidate_edges = self.G.get_edge_data(u, v)
            # There are many of these because this is a multidigraph.
            for _, edge in candidate_edges.items():
                if edge["exists"] == rel["exists"] and edge["action"] == rel["type"]:
                    # Don't need to re-add an identical edge
                    return
        if (u, v) not in self.edge_constraints:
            self.edge_constraints[(u, v)] = {}
        self.edge_constraints[(u, v)].update(attrs)
        self.G.add_edge(
            u, v, exists=rel["exists"], action=rel["type"], constraints=attrs
        )

    def automorphism_notation(self, tup):
        self.automorphisms.append(tup)

    def relation(self, tup):
        exists, typ = tup
        return {"exists": exists, "type": typ}

    def node_id(self, node_id):
        return str(node_id)

    def variable(self, var):
        return str("".join(str(c) for c in var))

    def rel_exist(self, _):
        return True

    def rel_nexist(self, _):
        return False

    def rel_def(self, _):
        return "SYN"

    def rel_neg(self, _):
        return "INH"

    def rel_pos(self, _):
        return "EXC"

    def iter_op_contains(self, _):
        return "contains"

    def iter_op_in(self, _):
        return "in"

    def iter_op_not_contains(self, _):
        return "!contains"

    def iter_op_not_in(self, _):
        return "!in"

    # Macros
    def macro(self, arg):
        name, args, rules = arg
        self.macros[name] = {"args": args, "rules": rules}

    def arglist(self, args):
        return [str(s) for s in args]

    def edge_macro(self, tup):
        if len(tup) == 3:
            u, rel, v = tup
            attrs = {}
        elif len(tup) == 4:
            u, rel, v, attrs = tup
        # else:
        #     print(tup, len(tup))
        u = str(u)
        v = str(v)
        return (str(u), rel, str(v), attrs)

    def macro_edge_clauses(self, rules):
        return list(rules)

    def macro_rules(self, rules):
        return list(rules)

    def macro_call(self, tup):
        callname, args = tup
        if callname not in self.macros:
            raise ValueError(
                f"Tried to invoke macro '{callname}' but "
                f"macro '{callname}' does not exist."
            )
        macro = self.macros[callname]
        macro_args = macro["args"]
        if isinstance(args, str):
            args = [args]
        if len(macro_args) != len(args):
            raise ValueError(
                f"Tried to invoke macro '{callname}' with "
                f"{len(args)} arguments, but '{callname}' takes "
                f"{len(macro_args)} arguments."
                f"\n"
                f"Called With:     {args}\n"
                f"Macro Arguments: {macro_args}"
            )

        # Else, append the macro to the graph:
        all_rules = []
        for rule in macro["rules"]:
            if isinstance(rule, tuple):
                all_rules.append(rule)
            else:
                for r in rule:
                    all_rules.append(r)

        for rule in all_rules:
            if len(rule) == 3:
                # This is a structural edge with no constraints.
                left, rel, right = rule
                attrs = {}
            elif len(rule) == 4:
                # This is an edge with attributes.
                left, rel, right, attrs = rule
            elif rule[0] == "node_constraint" and len(rule) == 5:
                # This is a node constraint!
                _, node, key, op, val = rule
                node = args[macro_args.index(node)]
                self.node_constraint((node, key, op, val))
                continue
            elif rule[0] == "node_constraint" and len(rule) == 6:
                # This is a dynamic node constraint!
                _, this_node, this_key, op, that_node, that_key = rule
                this_node = args[macro_args.index(this_node)]
                self.node_constraint((this_node, this_key, op, that_node, that_key))
                continue
            # Get the arguments in-place. For example, if left is A,
            # and A is the first arg in macro["args"], then replace
            # all instances of A in the rules with the first arg
            # from the macro call.
            left = str(left)
            right = str(right)
            left = args[macro_args.index(left)]
            right = args[macro_args.index(right)]
            dict_attrs = {}
            for k, v, a in attrs:
                if k not in dict_attrs:
                    dict_attrs[k] = {}
                if v in dict_attrs[k]:
                    dict_attrs[k][v].append(a)
                else:
                    dict_attrs[k][v] = [a]
            self.edge((left, rel, right, dict_attrs))

    def macro_call_re(self, tup):
        callname, args = tup
        if callname not in self.macros:
            raise ValueError(
                f"Tried to invoke macro '{callname}' but "
                f"macro {callname} does not exist."
            )
        macro = self.macros[callname]
        macro_args = macro["args"]
        if len(macro_args) != len(args):
            raise ValueError(
                f"Tried to invoke macro '{callname}' with "
                f"{len(args)} arguments, but {callname} takes "
                f"{len(macro_args)} arguments."
            )
        all_rules = []
        for rule in macro["rules"]:
            if isinstance(rule, tuple):
                all_rules.append(rule)
            else:
                for r in rule:
                    all_rules.append(r)
        return [
            (args[macro_args.index(rule[0])], rule[1], args[macro_args.index(rule[2])])
            for rule in all_rules
        ]


class ParserV2(Parser):
    def __init__(self, validators: List[Validator] = None) -> None:
        if validators is None:
            self.validators: List[Validator] = []
        else:
            self.validators = validators

    def parse(self, dm: str) -> nx.MultiDiGraph:
        """
        """
        G = nx.MultiDiGraph()

        tree = dm_parser.parse(dm)
        (
            G,
            edge_constraints,
            node_constraints,
            dynamic_edge_constraints,
            dynamic_node_constraints,
            automorphisms,
        ) = DotMotifTransformer(validators=self.validators).transform(tree)
        return (
            G,
            edge_constraints,
            node_constraints,
            dynamic_edge_constraints,
            dynamic_node_constraints,
            automorphisms,
        )




"""
Copyright 2022 The Johns Hopkins University Applied Physics Laboratory.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.`
"""

from itertools import product
import os
import time
from uuid import uuid4

from py2neo import Graph
import tamarind

# Types only:
import networkx as nx
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .. import dotmotif

from .Executor import Executor
from ..ingest import NetworkXIngester


def _remapped_operator(op):
    return {
        "=": "=",
        "==": "=",
        ">=": ">=",
        "<=": "<=",
        "<": "<",
        ">": ">",
        "!=": "<>",
        "<>": "<>",
        "in": "IN",
        "contains": "CONTAINS",
        "!in": "IN",
        "!contains": "CONTAINS",
    }[op]


def _operator_negation_infix(op):
    return {
        "=": False,
        "==": False,
        ">=": False,
        "<=": False,
        "<": False,
        ">": False,
        "!=": False,
        "<>": False,
        "in": False,
        "contains": False,
        "!in": True,
        "!contains": True,
    }[op]


def _quoted_if_necessary(val: str) -> str:
    """
    If the value is already quoted, return it.

    If it has single quotes in it, nest it in double quotes.
    If it has double quotes in it, nest it in single quotes.
    If it has both, nest it in double quotes and escape the double quotes in
    the existing text (i.e., [foo"bar] becomes ["foo\"bar"])
    """
    if val.startswith('"') and val.endswith('"'):
        return val
    elif val.startswith("'") and val.endswith("'"):
        return val
    elif '"' in val and "'" in val:
        return '"' + val.replace('"', '\\"') + '"'
    elif '"' in val:
        return "'" + val + "'"
    elif "'" in val:
        return '"' + val + '"'
    else:
        return '"' + val + '"'


_LOOKUP = {
    "INHIBITS": "INH",
    "EXCITES": "EXC",
    "SYNAPSES": "SYN",
    "INH": "INH",
    "EXC": "EXC",
    "SYN": "SYN",
    "DEFAULT": "SYN",
}

_DEFAULT_ENTITY_LABELS = {
    "node": "Neuron",
    "edge": _LOOKUP,
}


class Neo4jExecutor(Executor):
    """
    A Neo4j executor that runs Cypher queries against a running Neo4j database.

    .
    """

    def __init__(self, **kwargs) -> None:
        """
        Create a new executor.

        If there is an already-running Neo4j database, you can pass in
        authentication information and it will connect.

        If there is no existing database, you can pass in a graph to ingest
        and the executor will connect to it automatically.

        If there is no existing database and you do not pass in a graph, you
        must pass an `import_directory`, which the container will mount as an
        importable CSV resource.

        Arguments:
            db_bolt_uri (str): If connecting to an existing server, the URI
                of the server (including the port, probably 7474).
            username (str: "neo4j"): The username to use to attach to an
                existing server.
            password (str): The password to use to attach to an existing server.
            graph (nx.Graph): If provisioning a new database, the networkx
                graph to import into the database.
            import_directory (str): If provisioning a new database, the local
                directory to crawl for CSVs to import into the Neo4j database.
                Commonly used when you want to quickly and easily start a new
                Executor that uses the export from a previous graph.
            autoremove_container (bool: True): Whether to delete the container
                when the executor is deconstructed. Set to False if you'd like
                to be able to connect with other executors after the first one
                has closed.
            max_memory (str: "4G"): The maximum amount of memory to provision.
            initial_memory (str: "2G"): The starting heap-size for the Neo4j
                container's JVM.
            max_retries (int: 20): The number of times DotMotif should try to
                connect to the neo4j container before giving up.
            wait_for_boot (bool: True): Whether the process should pause to
                wait for a provisioned Docker container to come online.
            entity_labels (dict: _DEFAULT_ENTITY_LABELS): The set of labels to
                use for nodes and edges.

        """
        db_bolt_uri: Optional[str] = kwargs.get("db_bolt_uri", None)
        username: str = kwargs.get("username", "neo4j")
        password: Optional[str] = kwargs.get("password", None)
        self._autoremove_container: bool = kwargs.get("autoremove_container", True)
        self._wait_for_boot: bool = kwargs.get("wait_for_boot", True)
        self._max_memory_size: str = kwargs.get("max_memory", "4G")
        self._initial_heap_size: str = kwargs.get("initial_memory", "2G")
        self.max_retries: int = kwargs.get("max_retries", 20)
        self._entity_labels = kwargs.get("entity_labels", _DEFAULT_ENTITY_LABELS)

        graph: nx.Graph = kwargs.get("graph", None)
        import_directory: Optional[str] = kwargs.get("import_directory", None)

        self._created_container = False
        self._tamarind_provisioner = None

        if (
            (db_bolt_uri and graph)
            or (db_bolt_uri and import_directory)
            or (import_directory and graph)
        ):
            raise ValueError(
                "Specify EXACTLY ONE of db_bolt_uri/graph/import_directory."
            )

        if db_bolt_uri:
            # Authentication information was provided. Use this to log in and
            # connect to the existing database.
            self._connect_to_existing_graph(db_bolt_uri, username, password)

        elif graph:
            export_dir = "export-custom-graph"
            # A networkx graph was provided.
            # We must export this to a set of CSV files, drop them to disk,
            # and then we can use the same strategy as `import_directory` to
            # run a container.
            nxi = NetworkXIngester(graph, export_dir)
            try:
                nxi.ingest()
            except Exception as e:
                raise ValueError(f"Could not export graph: {e}")

            self._tamarind_provisioner = tamarind.Neo4jDockerProvisioner(
                autoremove_containers=self._autoremove_container,
                max_memory_size=self._max_memory_size,
                initial_heap_size=self._initial_heap_size,
            )
            self._create_container(export_dir)

        elif import_directory:
            self._tamarind_provisioner = tamarind.Neo4jDockerProvisioner(
                autoremove_containers=self._autoremove_container,
                max_memory_size=self._max_memory_size,
                initial_heap_size=self._initial_heap_size,
            )
            self._create_container(import_directory)

        else:
            raise ValueError(
                "You must supply either an existing db or a graph to load."
            )

    def __del__(self):
        """
        Destroy the docker container from the running processes.

        Also will handle (TODO) other teardown actions.
        """
        if self._created_container:
            self._teardown_container()

    def _connect_to_existing_graph(
        self, db_bolt_uri: str, username: str, password: str
    ) -> None:
        try:
            self.G = Graph(db_bolt_uri, username=username, password=password)
        except:
            raise ValueError(f"Could not connect to graph {db_bolt_uri}.")

    def _create_container(self, import_dir: str):
        # Create a docker container:

        _run_before = (
            f"""./bin/neo4j-admin import --id-type STRING --nodes={self._entity_labels['node']}="""
            + f""""/import/export-neurons-.*.csv" --relationships={self._entity_labels['edge']['DEFAULT']}="""
            + """"/import/export-synapses-.*.csv" """
        )

        self._tamarind_container_id = str(uuid4())
        (
            self._running_container,
            self._container_port,
        ) = self._tamarind_provisioner.start(
            self._tamarind_container_id,
            import_path=f"{os.getcwd()}/{import_dir}",
            run_before=_run_before,
            wait=self._wait_for_boot,
            wait_attempt_limit=self.max_retries,
        )
        self._created_container = True
        container_is_ready = False
        tries = 0
        while not container_is_ready:
            try:
                self.G = self._tamarind_provisioner[self._tamarind_container_id]
                container_is_ready = True
            except Exception as e:
                tries += 1
                if tries > self.max_retries:
                    raise IOError(
                        f"Could not connect to neo4j container {self._running_container}. "
                        "For more information, see https://github.com/aplbrain/dotmotif/wiki/Troubleshooting-Neo4jExecutor."
                    )
                time.sleep(3)
        self.G = self._tamarind_provisioner[self._tamarind_container_id]

    def _teardown_container(self):
        self._tamarind_provisioner.stop(self._tamarind_container_id)

    def create_index(self, attribute_name: str):
        """
        Create a new index on the given node attribute.

        Note that edge attributes are NOT supported.
        """
        self.run(
            f"""
        CREATE INDEX index_name IF NOT EXISTS FOR (n:{self._entity_labels["node"]})
        ON (n.{attribute_name})
        """
        )

    def run(self, cypher: str, cursor=True):
        """
        Run an arbitrary cypher command.

        You should usually ignore this, and use .find() instead.

        Arguments:
            cypher (str): The command to run

        Returns:
            The result of the cypher query (py2neo.Table)

        """
        if not cursor:
            return self.G.run(cypher).to_table()
        return self.G.run(cypher)

    def count(self, motif: "dotmotif.Motif", limit=None) -> int:
        """
        Count a motif in a larger graph.

        Arguments:
            motif (dotmotif.Motif)

        """
        qry = self.motif_to_cypher(
            motif, count_only=True, static_entity_labels=self._entity_labels
        )
        if limit:
            qry += f" LIMIT {limit}"
        return int(self.G.run(qry).to_ndarray())

    def find(self, motif: "dotmotif.Motif", limit=None, cursor=True):
        """
        Find a motif in a larger graph.

        Arguments:
            motif (dotmotif.Motif)

        """
        qry = self.motif_to_cypher(motif, static_entity_labels=self._entity_labels)
        if limit:
            qry += f" LIMIT {limit}"
        if not cursor:
            return self.G.run(qry).to_table()
        return self.G.run(qry)

    @staticmethod
    def motif_to_cypher(
        motif: "dotmotif.Motif",
        count_only: bool = False,
        static_entity_labels: dict = None,
    ) -> str:
        """
        Output a query suitable for Cypher-compatible engines (e.g. Neo4j).

        Returns:
            str: A Cypher query

        """
        static_entity_labels = static_entity_labels or _DEFAULT_ENTITY_LABELS
        # Edges and negative edges
        es = []
        es_neg = []

        motif_graph = motif.to_nx()

        # ID that is assigned to it so that it can hold constraints later on.
        edge_mapping = {}

        for u, v, a in motif_graph.edges(data=True):
            action = static_entity_labels["edge"][
                a.get("action", static_entity_labels["edge"]["DEFAULT"])
            ]
            edge_id = "{}_{}".format(u, v)
            edge_mapping[(u, v)] = edge_id
            if a["exists"]:
                es.append(
                    (
                        "MATCH ({}"
                        + (
                            (":" + static_entity_labels["node"])
                            if static_entity_labels["node"]
                            else ""
                        )
                        + ")-[{}{}]-{}({}"
                        + (
                            (":" + static_entity_labels["node"])
                            if static_entity_labels["node"]
                            else ""
                        )
                        + ")"
                    ).format(
                        u,
                        edge_id,
                        ((":" + action) if action else ""),
                        "" if motif.ignore_direction else ">",
                        v,
                    )
                )
            else:
                es_neg.append(
                    (
                        "NOT ({}"
                        + (
                            (":" + static_entity_labels["node"])
                            if static_entity_labels["node"]
                            else ""
                        )
                        + ")-[:{}]-{}({}"
                        + (
                            (":" + static_entity_labels["node"])
                            if static_entity_labels["node"]
                            else ""
                        )
                        + ")"
                    ).format(u, action, "" if motif.ignore_direction else ">", v)
                )

        delim = "\n" if motif.pretty_print else " "

        conditions = []
        if es_neg:
            conditions = es_neg  # f"{delim} AND ".join(es_neg)

        q_match = delim.join([delim.join(es)])

        # Edge constraints:
        cypher_edge_constraints = []
        for (u, v), a in motif.list_edge_constraints().items():
            for key, constraints in a.items():
                for operator, values in constraints.items():
                    for value in values:
                        cypher_edge_constraints.append(
                            (
                                "NOT ({}[{}] {} {})"
                                if _operator_negation_infix(operator)
                                else "{}[{}] {} {}"
                            ).format(
                                edge_mapping[(u, v)],
                                _quoted_if_necessary(key),
                                _remapped_operator(operator),
                                f'"{value}"' if isinstance(value, str) else value,
                            )
                        )

        # Node constraints:
        cypher_node_constraints = []
        for n, a in motif.list_node_constraints().items():
            for key, constraints in a.items():
                for operator, values in constraints.items():
                    for value in values:
                        cypher_node_constraints.append(
                            (
                                "NOT ({}[{}] {} {})"
                                if _operator_negation_infix(operator)
                                else "{}[{}] {} {}"
                            ).format(
                                n,
                                _quoted_if_necessary(key),
                                _remapped_operator(operator),
                                f'"{value}"' if isinstance(value, str) else value,
                            )
                        )

        # Dynamic node constraints:
        for n, a in motif.list_dynamic_node_constraints().items():
            for key, constraints in a.items():
                for operator, values in constraints.items():
                    for value in values:
                        cypher_node_constraints.append(
                            (
                                "NOT ({}[{}] {} {}[{}])"
                                if _operator_negation_infix(operator)
                                else "{}[{}] {} {}[{}]"
                            ).format(
                                n,
                                _quoted_if_necessary(key),
                                _remapped_operator(operator),
                                value[0],
                                _quoted_if_necessary(value[1]),
                            )
                        )

        conditions.extend([*cypher_node_constraints, *cypher_edge_constraints])

        if count_only:
            q_return = (
                "WITH DISTINCT "
                + ",".join(list(motif_graph.nodes()))
                + " AS __DOTMOTIF_DISTINCT "
                + delim
                + "RETURN COUNT(*)"
            )
        else:
            q_return = "RETURN DISTINCT " + ",".join(list(motif_graph.nodes()))

        if motif.limit:
            q_limit = " LIMIT {}".format(motif.limit)
        else:
            q_limit = ""

        if motif.enforce_inequality:
            _nodes = [str(a) for a in motif_graph.nodes()]
            conditions.extend(
                sorted(
                    list(
                        {
                            "<>".join(sorted(a))
                            for a in list(product(_nodes, _nodes))
                            if a[0] != a[1]
                        }
                    )
                )
            )

        automs = motif.list_automorphisms()
        conditions.extend(["id({}) < id({})".format(a, b) for a, b in automs])

        query = [q_match]
        if conditions:
            query.append("WHERE " + " AND ".join(conditions))
        query.append(q_return)
        if q_limit:
            query.append(q_limit)
        return "{}".format(delim.join(query))



import dotmotif

import unittest

_THREE_CYCLE = """A -> B\nB -> C\nC -> A\n"""
_THREE_CYCLE_NEG = """A !> B\nB !> C\nC !> A\n"""
_THREE_CYCLE_INH = """A -| B\nB -| C\nC -| A\n"""
_THREE_CYCLE_NEG_INH = """A !| B\nB !| C\nC !| A\n"""

_SEMICOLON_TRIANGLE = """A->B;B->C;C->A"""


class TestDotmotif_Parserv2_DM(unittest.TestCase):
    def test_sanity(self):
        self.assertEqual(1, 1)

    def test_dm_parser(self):
        dm = dotmotif.Motif(_THREE_CYCLE)
        self.assertEqual(len(dm._g.edges()), 3)
        self.assertEqual(len(dm._g.nodes()), 3)

    def test_dm_parser_with_semicolons(self):
        dm = dotmotif.Motif(_SEMICOLON_TRIANGLE)
        self.assertEqual(len(dm._g.edges()), 3)
        self.assertEqual(len(dm._g.nodes()), 3)

    def test_dm_parser_actions(self):
        dm = dotmotif.Motif(_THREE_CYCLE)
        self.assertEqual([e[2]["action"] for e in dm._g.edges(data=True)], ["SYN"] * 3)

        dm = dotmotif.Motif(_THREE_CYCLE_INH)
        self.assertEqual([e[2]["action"] for e in dm._g.edges(data=True)], ["INH"] * 3)

    def test_dm_parser_edge_exists(self):
        dm = dotmotif.Motif(_THREE_CYCLE)
        self.assertEqual([e[2]["exists"] for e in dm._g.edges(data=True)], [True] * 3)

        dm = dotmotif.Motif(_THREE_CYCLE_NEG)
        self.assertEqual([e[2]["exists"] for e in dm._g.edges(data=True)], [False] * 3)

        dm = dotmotif.Motif(_THREE_CYCLE_NEG_INH)
        self.assertEqual([e[2]["exists"] for e in dm._g.edges(data=True)], [False] * 3)


class TestDotmotif_Parserv2_DM_Macros(unittest.TestCase):
    def test_macro_not_added(self):
        exp = """\
        edge(A, B) {
            A -> B
        }
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm._g.edges()), 0)

    def test_simple_macro(self):
        exp = """\
        edge(A, B) {
            A -> B
        }
        edge(C, D)
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm._g.edges()), 1)

    def test_simple_macro_construction(self):
        exp = """\
        edge(A, B) {
            A -> B
        }
        edge(C, D)
        """
        dm = dotmotif.Motif(exp)
        exp_edge = list(dm._g.edges(data=True))[0]
        self.assertEqual(exp_edge[0], "C")
        self.assertEqual(exp_edge[1], "D")

    def test_multiline_macro_construction(self):
        exp = """\
        dualedge(A, B) {
            A -> B
            B -> A
        }
        dualedge(C, D)
        """
        dm = dotmotif.Motif(exp)
        exp_edge = list(dm._g.edges(data=True))[0]
        self.assertEqual(exp_edge[0], "C")
        self.assertEqual(exp_edge[1], "D")

    def test_undefined_macro(self):
        exp = """\
        dualedge(A, B) {
            A -> B
            B -> A
        }
        foo(C, D)
        """
        # with self.assertRaises(ValueError):
        with self.assertRaises(Exception):
            dotmotif.Motif(exp)

    def test_wrong_args_macro(self):
        exp = """\
        edge(A, B) {
            A -> B
            B -> A
        }
        edge(C, D, E)
        """
        # with self.assertRaises(ValueError):
        with self.assertRaises(Exception):
            dotmotif.Motif(exp)

    def test_more_complex_macro(self):
        exp = """\
        tri(A, B, C) {
            A -> B
            B -> C
            C -> A
        }
        tri(C, D, E)
        """
        dm = dotmotif.Motif(exp)
        edges = list(dm._g.edges(data=True))
        self.assertEqual(len(edges), 3)

    def test_macro_reuse(self):
        exp = """\
        tri(A, B, C) {
            A -> B
            B -> C
            C -> A
        }
        tri(C, D, E)
        tri(F, G, H)
        """
        dm = dotmotif.Motif(exp)
        edges = list(dm._g.edges(data=True))
        self.assertEqual(len(edges), 6)

    def test_conflicting_macro_invalid_edge_throws(self):
        exp = """\
        tri(A, B, C) {
            A -> B
            B -> C
            C -> A
        }

        nontri(A, B, C) {
            A !> B
            B !> C
            C !> A
        }
        tri(C, D, E)
        nontri(D, E, F)
        """
        # with self.assertRaises(dotmotif.validators.DisagreeingEdgesValidatorError):
        with self.assertRaises(Exception):
            dotmotif.Motif(exp)

    def test_nested_macros(self):
        exp = """\
        dualedge(A, B) {
            A -> B
            B -> A
        }
        dualtri(A, B, C) {
            dualedge(A, B)
            dualedge(B, C)
            dualedge(C, A)
        }
        dualtri(foo, bar, baz)
        """
        dm = dotmotif.Motif(exp)
        edges = list(dm._g.edges(data=True))
        self.assertEqual(len(edges), 6)

    def test_deeply_nested_macros(self):
        exp = """\
        edge(A, B) {
            A -> B
        }
        dualedge(A, B) {
            edge(A, B)
            edge(B, A)
        }
        dualtri(A, B, C) {
            dualedge(A, B)
            dualedge(B, C)
            dualedge(C, A)
        }
        dualtri(foo, bar, baz)
        """
        dm = dotmotif.Motif(exp)
        edges = list(dm._g.edges(data=True))
        self.assertEqual(len(edges), 6)

    def test_clustercuss_macros_no_repeats(self):
        exp = """\
        edge(A, B) {
            A -> B
        }
        dualedge(A, B) {
            edge(A, B)
            edge(B, A)
        }

        dualtri(A, B, C) {
            dualedge(A, B)
            dualedge(B, C)
            dualedge(C, A)
        }

        dualtri(foo, bar, baz)
        dualtri(foo, bar, baf)
        """
        dm = dotmotif.Motif(exp)
        edges = list(dm._g.edges(data=True))
        self.assertEqual(len(edges), 10)

    def test_comment_in_macro(self):
        exp = """\
        # Outside comment
        edge(A, B) {
            # Inside comment
            A -> B
        }
        dualedge(A, B) {
            # Nested-inside comment
            edge(A, B)
            edge(B, A)
        }

        dualedge(foo, bar)
        """
        dm = dotmotif.Motif(exp)
        edges = list(dm._g.edges(data=True))
        self.assertEqual(len(edges), 2)

    def test_combo_macro(self):
        exp = """\
        edge(A, B) {
            A -> B
        }
        dualedge(A, B) {
            # Nested-inside comment!
            edge(A, B)
            B -> A
        }

        dualedge(foo, bar)
        """
        dm = dotmotif.Motif(exp)
        edges = list(dm._g.edges(data=True))
        self.assertEqual(len(edges), 2)

    def test_comment_macro_inline(self):
        exp = """\
        edge(A, B) {
            A -> B
        }
        dualedge(A, B) {
            # Nested-inside comment!
            edge(A, B) # inline comment
            B -> A
        }

        dualedge(foo, bar) # inline comment
        # standalone comment
        foo -> bar # inline comment
        """
        dm = dotmotif.Motif(exp)
        edges = list(dm._g.edges(data=True))
        self.assertEqual(len(edges), 2)

    def test_alphanumeric_variables(self):
        exp = """\
        edge(A, B) {
            A -> B
        }
        dualedge(A1, B) {
            # Nested-inside comment!
            edge(A1, B) # inline comment
            B -> A1
        }

        dualedge(foo_1, bar_2) # inline comment
        # standalone comment
        foo_1 -> bar_2 # inline comment
        """
        dm = dotmotif.Motif(exp)
        edges = list(dm._g.edges(data=True))
        self.assertEqual(len(edges), 2)
        self.assertEqual(list(dm._g.nodes()), ["foo_1", "bar_2"])
        self.assertEqual(type(list(dm._g.nodes())[0]), str)

        new_exp = """
        L1 -> Mi1
        L1 -> Tm3
        L3 -> Mi9
        """
        dm = dotmotif.Motif(new_exp)
        self.assertEqual(list(dm._g.nodes()), ["L1", "Mi1", "Tm3", "L3", "Mi9"])


class TestDotmotif_Parserv2_DM_EdgeAttributes(unittest.TestCase):
    def test_basic_edge_attr(self):
        exp = """\
        Aa -> Ba [type == 1]
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm._g.edges()), 1)
        u, v, d = list(dm._g.edges(["Aa", "Bb"], data=True))[0]
        self.assertEqual(type(list(dm._g.nodes())[0]), str)
        self.assertEqual(type(list(dm._g.nodes())[1]), str)
        self.assertEqual(d["constraints"]["type"], {"==": [1]})

    def test_edge_multi_attr(self):
        exp = """\
        Aa -> Ba [type != 1, type != 12]
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm._g.edges()), 1)
        u, v, d = list(dm._g.edges(data=True))[0]
        self.assertEqual(d["constraints"]["type"], {"!=": [1, 12]})

    def test_edge_macro_attr(self):
        exp = """\
        macro(Aa, Ba) {
            Aa -> Ba [type != 1, type != 12]
        }

        macro(X, Y)
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm._g.edges()), 1)
        u, v, d = list(dm._g.edges(data=True))[0]
        self.assertEqual(d["constraints"]["type"], {"!=": [1, 12]})


class TestDotmotif_Parserv2_DM_NodeAttributes(unittest.TestCase):
    def test_basic_node_attr(self):
        exp = """\
        Aa -> Ba

        Aa.type = "excitatory"
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm.list_node_constraints()), 1)
        self.assertEqual(list(dm.list_node_constraints().keys()), ["Aa"])

    def test_basic_node_attr_bracket_keying(self):
        exp = """\
        Aa -> Ba

        Aa['type'] = "excitatory"
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm.list_node_constraints()), 1)
        self.assertEqual(list(dm.list_node_constraints().keys()), ["Aa"])

    def test_node_multi_attr(self):
        exp = """\
        Aa -> Ba

        Aa.type = "excitatory"
        Aa.size = 4.5
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm.list_node_constraints()), 1)
        self.assertEqual(len(dm.list_node_constraints()["Aa"]), 2)
        self.assertEqual(dm.list_node_constraints()["Aa"]["type"]["="], ["excitatory"])
        self.assertEqual(dm.list_node_constraints()["Aa"]["size"]["="], [4.5])
        self.assertEqual(list(dm.list_node_constraints().keys()), ["Aa"])

    def test_multi_node_attr(self):
        exp = """\
        Aa -> Ba

        Aa.type = "excitatory"
        Ba.size=4.0
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm.list_node_constraints()), 2)
        self.assertEqual(list(dm.list_node_constraints().keys()), ["Aa", "Ba"])

    def test_node_macro_attr(self):
        exp = """\
        macro(A) {
            A.type = "excitatory"
            A.size >= 4.0
        }
        Aaa -> Ba
        macro(Aaa)
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm.list_node_constraints()), 1)
        self.assertEqual(list(dm.list_node_constraints().keys()), ["Aaa"])

        exp = """\
        macro(A) {
            A.type = "excitatory"
            A.size >= 4.0
        }
        Aaa -> Ba
        macro(Aaa)
        macro(Ba)
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm.list_node_constraints()), 2)
        self.assertEqual(list(dm.list_node_constraints().keys()), ["Aaa", "Ba"])


class TestDynamicNodeConstraints(unittest.TestCase):
    def test_dynamic_constraints(self):
        """
        Test that comparisons may be made between variables, e.g.:

        A.type != B.type

        """
        exp = """\
        A -> B
        A.radius < B.radius
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm.list_dynamic_node_constraints()), 1)

    def test_dynamic_constraints_brackets(self):
        """
        Test that comparisons may be made between variables, e.g.:

        A.type != B.type

        """
        exp = """\
        A -> B
        A['radius'] < B["radius"]
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm.list_dynamic_node_constraints()), 1)

    def test_dynamic_constraints_in_macro(self):
        """
        Test that comparisons may be made between variables in a macro, e.g.:

        A.type != B.type

        """
        exp = """\
        macro(A, B) {
            A.radius > B.radius
        }
        macro(A, B)
        A -> B
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm.list_dynamic_node_constraints()), 1)

    def test_dynamic_constraints_bracketed_in_macro(self):
        """
        Test that comparisons may be made between variables in a macro, e.g.:

        A.type != B.type

        """
        exp = """\
        macro(A, B) {
            # A.radius > B.radius
            A["radius"] > B['radius']
        }
        macro(A, B)
        A -> B
        """
        dm = dotmotif.Motif(exp)
        self.assertEqual(len(dm.list_dynamic_node_constraints()), 1)



import unittest
import dotmotif
from dotmotif import Motif
from dotmotif.executors import GrandIsoExecutor
from dotmotif.executors.NetworkXExecutor import (
    _edge_satisfies_constraints,
    _node_satisfies_constraints,
)
from dotmotif.parsers.v2 import ParserV2
import networkx as nx


class TestSmallMotifs(unittest.TestCase):
    def test_edgecount_motif(self):
        dm = Motif("""A->B""")

        H = nx.DiGraph()
        H.add_edge("x", "y")
        self.assertEqual(len(GrandIsoExecutor(graph=H).find(dm)), 1)
        H.add_edge("x", "y")
        self.assertEqual(len(GrandIsoExecutor(graph=H).find(dm)), 1)
        H.add_edge("x", "z")
        self.assertEqual(len(GrandIsoExecutor(graph=H).find(dm)), 2)

    def test_fullyconnected_triangle_motif(self):
        dm = Motif(
            """
        A->B
        B->C
        C->A
        """
        )

        H = nx.DiGraph()
        H.add_edge("x", "y")
        self.assertEqual(len(GrandIsoExecutor(graph=H).find(dm)), 0)
        H.add_edge("y", "z")
        self.assertEqual(len(GrandIsoExecutor(graph=H).find(dm)), 0)
        H.add_edge("z", "x")
        self.assertEqual(len(GrandIsoExecutor(graph=H).find(dm)), 3)

    def test_edge_attribute_equality(self):
        dm = Motif(
            """
        A->B [weight==10, area==4]
        """
        )

        H = nx.DiGraph()
        H.add_edge("z", "x", weight=10, area=4)
        H.add_edge("x", "y")
        H.add_edge("y", "z", weight=5)
        self.assertEqual(len(GrandIsoExecutor(graph=H).find(dm)), 1)

    def test_one_instance(self):

        H = nx.DiGraph()
        H.add_edge("x", "y", weight=1)
        H.add_edge("y", "z", weight=10)
        H.add_edge("z", "x", weight=5)
        motif = dotmotif.Motif(
            """
        A -> B [weight>=11]
        """.strip()
        )

        self.assertEqual(len(GrandIsoExecutor(graph=H).find(motif)), 0)

    def test_two_instance(self):
        H = nx.DiGraph()
        H.add_edge("x", "y", weight=1)
        H.add_edge("y", "z", weight=10)
        H.add_edge("z", "x", weight=5)
        H.add_edge("z", "a", weight=5)
        H.add_edge("a", "b", weight=1)
        H.add_edge("b", "c", weight=10)
        H.add_edge("c", "a", weight=5)
        motif = dotmotif.Motif(
            """
        A -> B [weight>=7]
        """.strip()
        )

        self.assertEqual(len(GrandIsoExecutor(graph=H).find(motif)), 2)

    def test_triangle_two_instance(self):
        H = nx.DiGraph()
        H.add_edge("x", "y", weight=1)
        H.add_edge("y", "z", weight=10)
        H.add_edge("z", "x", weight=5)
        H.add_edge("z", "a", weight=5)
        H.add_edge("a", "b", weight=1)
        H.add_edge("b", "c", weight=10)
        H.add_edge("c", "a", weight=5)
        motif = dotmotif.Motif(
            """
        A -> B [weight>=7]
        B -> C
        C -> A
        """.strip()
        )

        self.assertEqual(len(GrandIsoExecutor(graph=H).find(motif)), 2)

    def test_mini_example(self):

        H = nx.DiGraph()
        H.add_edge("y", "x", ATTRIBUTE=7)
        H.add_edge("y", "z", ATTRIBUTE=7)
        motif = dotmotif.Motif(
            """
        A -> B [ATTRIBUTE>=7]
        """.strip()
        )

        self.assertEqual(len(GrandIsoExecutor(graph=H).find(motif)), 2)

    def test_node_and_edge_full_example(self):

        H = nx.DiGraph()
        H.add_edge("X", "Y", weight=10)
        H.add_edge("Y", "Z", weight=9)
        H.add_edge("Z", "X", weight=8)
        motif = dotmotif.Motif(
            """
            A -> B [weight>=7]
        """.strip()
        )

        res = GrandIsoExecutor(graph=H).find(motif)
        self.assertEqual(len(res), 3)

        H.add_edge("Z", "C", weight=7)
        res = GrandIsoExecutor(graph=H).find(motif)
        self.assertEqual(len(res), 4)

        H.add_edge("Z", "D", weight="no")
        res = GrandIsoExecutor(graph=H).find(motif)
        self.assertEqual(len(res), 4)

        H.add_edge("y", "a")
        self.assertEqual(len(GrandIsoExecutor(graph=H).find(motif)), 4)

        H.add_edge("y", "a", other_weight=7, weight=8)
        self.assertEqual(len(GrandIsoExecutor(graph=H).find(motif)), 5)

    def test_automorphism_reduction(self):

        G = nx.DiGraph()
        G.add_edge("X", "Z")
        G.add_edge("Y", "Z")

        motif = dotmotif.Motif(
            """
            A -> C
            B -> C

            A === B
            """
        )

        res = GrandIsoExecutor(graph=G).find(motif)
        self.assertEqual(len(res), 2)

        motif = dotmotif.Motif(exclude_automorphisms=True).from_motif(
            """
            A -> C
            B -> C

            A === B
            """
        )

        res = GrandIsoExecutor(graph=G).find(motif)
        self.assertEqual(len(res), 1)

    def test_automorphism_auto(self):

        G = nx.DiGraph()
        G.add_edge("X", "Z")
        G.add_edge("Y", "Z")

        motif = dotmotif.Motif(exclude_automorphisms=True).from_motif(
            """
            A -> C
            B -> C
            """
        )

        res = GrandIsoExecutor(graph=G).find(motif)
        self.assertEqual(len(res), 1)

    def test_automorphism_notauto(self):

        G = nx.DiGraph()
        G.add_edge("X", "Z")
        G.add_edge("Y", "Z")

        motif = dotmotif.Motif(
            """
            A -> C
            B -> C
            """
        )

        res = GrandIsoExecutor(graph=G).find(motif)
        self.assertEqual(len(res), 2)

    def test_automorphism_flag_triangle(self):

        G = nx.DiGraph()
        G.add_edge("A", "B")
        G.add_edge("B", "C")
        G.add_edge("C", "A")

        motif = dotmotif.Motif(
            """
            A -> B
            B -> C
            C -> A
            """
        )
        res = GrandIsoExecutor(graph=G).find(motif)
        self.assertEqual(len(res), 3)

        motif = dotmotif.Motif(exclude_automorphisms=True).from_motif(
            """
            A -> B
            B -> C
            C -> A
            """
        )
        res = GrandIsoExecutor(graph=G).find(motif)
        self.assertEqual(len(res), 1)


class TestDynamicNodeConstraints(unittest.TestCase):
    def test_dynamic_constraints_zero_results(self):
        """
        Test that comparisons may be made between variables, e.g.:

        A.type != B.type

        """
        G = nx.DiGraph()
        G.add_edge("A", "B")
        G.add_edge("B", "C")
        G.add_edge("C", "A")
        G.add_node("A", radius=5)
        G.add_node("B", radius=10)
        exp = """\
        A -> B
        A.radius > B.radius
        """
        dm = dotmotif.Motif(parser=ParserV2)
        res = GrandIsoExecutor(graph=G).find(dm.from_motif(exp))
        self.assertEqual(len(res), 0)

    def test_dynamic_constraints_one_result(self):
        """
        Test that comparisons may be made between variables, e.g.:

        A.type != B.type

        """
        G = nx.DiGraph()
        G.add_edge("A", "B")
        G.add_edge("B", "C")
        G.add_edge("C", "A")
        G.add_node("A", radius=25)
        G.add_node("B", radius=10)
        exp = """\
        A -> B
        A.radius > B.radius
        """
        dm = dotmotif.Motif(parser=ParserV2)
        res = GrandIsoExecutor(graph=G).find(dm.from_motif(exp))
        self.assertEqual(len(res), 1)

    def test_dynamic_constraints_two_results(self):
        """
        Test that comparisons may be made between variables, e.g.:

        A.type != B.type

        """
        G = nx.DiGraph()
        G.add_edge("A", "B")
        G.add_edge("B", "C")
        G.add_edge("C", "A")
        G.add_node("A", radius=25)
        G.add_node("B", radius=10)
        G.add_node("C", radius=5)
        exp = """\
        A -> B
        A.radius > B.radius
        """
        dm = dotmotif.Motif(parser=ParserV2)
        res = GrandIsoExecutor(graph=G).find(dm.from_motif(exp))
        self.assertEqual(len(res), 2)

    def test_dynamic_constraints_in_macros_zero_results(self):
        """
        Test that comparisons may be made between variables, e.g.:

        A.type != B.type

        """
        G = nx.DiGraph()
        G.add_edge("A", "B")
        G.add_edge("B", "C")
        G.add_edge("C", "A")
        G.add_node("A", radius=5)
        G.add_node("B", radius=10)
        exp = """\
        macro(A, B) {
            A.radius > B.radius
        }
        macro(A, B)
        A -> B
        """
        dm = dotmotif.Motif(parser=ParserV2)
        res = GrandIsoExecutor(graph=G).find(dm.from_motif(exp))
        self.assertEqual(len(res), 0)

    def test_dynamic_constraints_in_macros_one_result(self):
        """
        Test that comparisons may be made between variables, e.g.:

        A.type != B.type

        """
        G = nx.DiGraph()
        G.add_edge("A", "B")
        G.add_edge("B", "C")
        G.add_edge("C", "A")
        G.add_node("A", radius=15)
        G.add_node("B", radius=10)
        exp = """\
        macro(A, B) {
            A.radius > B.radius
        }
        macro(A, B)
        A -> B
        """
        dm = dotmotif.Motif(parser=ParserV2)
        res = GrandIsoExecutor(graph=G).find(dm.from_motif(exp))
        self.assertEqual(len(res), 1)

    def test_dynamic_constraints_in_macros_two_result(self):
        """
        Test that comparisons may be made between variables, e.g.:

        A.type != B.type

        """
        G = nx.DiGraph()
        G.add_edge("A", "B")
        G.add_edge("B", "C")
        G.add_edge("C", "A")
        G.add_node("A", radius=15)
        G.add_node("B", radius=10)
        G.add_node("C", radius=10)
        exp = """\
        macro(A, B) {
            A.radius >= B.radius
        }
        macro(A, B)
        A -> B
        """
        dm = dotmotif.Motif(parser=ParserV2)
        res = GrandIsoExecutor(graph=G).find(dm.from_motif(exp))
        self.assertEqual(len(res), 2)



"""
Copyright 2021 The Johns Hopkins University Applied Physics Laboratory.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.`
"""
from functools import lru_cache
import networkx as nx
from grandiso import find_motifs_iter

from .NetworkXExecutor import NetworkXExecutor, _node_satisfies_constraints


class GrandIsoExecutor(NetworkXExecutor):
    """
    A DotMotif executor that uses grandiso for subgraph monomorphism.

    This executor is dramatically fast than the NetworkX search, and is still
    a pure-Python implementation.

    [GrandIso](https://github.com/aplbrain/grandiso-networkx)

    """

    def find(self, motif, limit: int = None):
        """
        Find a motif in a larger graph.

        Arguments:
            motif (dotmotif.Motif)
            limit (int: None)

        Returns:
            List[dict]

        """
        # We need to first remove "negative" nodes from the motif, and then
        # filter them out later on.

        if motif.ignore_direction or not self.graph.is_directed:
            graph_constructor = nx.Graph
        else:
            graph_constructor = nx.DiGraph

        only_positive_edges_motif = graph_constructor()
        must_not_exist_edges = []
        for u, v, attrs in motif.to_nx().edges(data=True):
            if attrs["exists"] is True:
                only_positive_edges_motif.add_edge(u, v, **attrs)
            elif attrs["exists"] is False:
                # Collect a list of neg-edges to check for again in a moment
                must_not_exist_edges.append((u, v))

        def _doesnt_have_any_of_motifs_negative_edges(mapping):
            for u, v in must_not_exist_edges:
                if self.graph.has_edge(mapping[u], mapping[v]):
                    return False
            return True

        constraints = motif.list_node_constraints()

        @lru_cache()
        def _node_attr_match_fn(
            motif_node_id: str, host_node_id: str, motif_nx: nx.Graph, host_nx: nx.Graph
        ):
            return _node_satisfies_constraints(
                host_nx.nodes[host_node_id], constraints.get(motif_node_id, {})
            )
            return True

        graph_matches = find_motifs_iter(
            only_positive_edges_motif,
            self.graph,
            is_node_attr_match=_node_attr_match_fn,
            is_edge_attr_match=lambda _1, _2, _3, _4: True,
        )

        _edge_constraint_validator = (
            self._validate_edge_constraints
            if not self._host_is_multigraph
            else (
                self._validate_multigraph_all_edge_constraints
                if self._multigraph_edge_match == "all"
                else self._validate_multigraph_any_edge_constraints
            )
        )

        results = []
        for r in graph_matches:
            if _doesnt_have_any_of_motifs_negative_edges(r) and (
                _edge_constraint_validator(r, self.graph, motif.list_edge_constraints())
                # and self._validate_node_constraints(
                #     r, self.graph, motif.list_node_constraints()
                # )
                and self._validate_dynamic_node_constraints(
                    r, self.graph, motif.list_dynamic_node_constraints()
                )
                # by default, networkx returns the automorphism that is left-
                # sorted, so this comparison is _opposite_ the check that we
                # use in the other executors. In other words, we usually check
                # that A >= B; here we check A <= B.
                and (
                    (not motif.exclude_automorphisms)
                    or all(r[a] <= r[b] for (a, b) in motif.list_automorphisms())
                )
            ):
                results.append(r)
                if limit and len(results) >= limit:
                    return results

        return results



import dotmotif
from dotmotif.executors.Neo4jExecutor import Neo4jExecutor, _quoted_if_necessary
import unittest
import networkx as nx


class TestNeo4jExecutor(unittest.TestCase):
    def test_numerical_ids_fail_with_message_github_26(self):
        g = nx.DiGraph()
        g.add_edge(1, 2)
        with self.assertRaisesRegex(Exception, "numerical IDs"):
            Neo4jExecutor(graph=g)


class TestNeo4jExecutor_Automorphisms(unittest.TestCase):
    def test_basic_node_attr(self):
        exp = """\
        A -> C
        B -> C
        A === B
        """
        dm = dotmotif.Motif(exp)
        cypher = Neo4jExecutor.motif_to_cypher(dm)
        self.assertIn("id(A) < id(B)", cypher)

    def test_automatic_autos(self):
        exp = """\
        A -> C
        B -> C
        """
        dm = dotmotif.Motif(exp, exclude_automorphisms=True)
        cypher = Neo4jExecutor.motif_to_cypher(dm)
        self.assertIn("id(A) < id(B)", cypher)


class TestQuotingIfNecessary(unittest.TestCase):
    def test_quoting_if_necessary(self):
        self.assertEqual(_quoted_if_necessary('"xyz"'), '"xyz"'),
        self.assertEqual(_quoted_if_necessary("'xyz'"), "'xyz'"),
        self.assertEqual(_quoted_if_necessary("""x'y"z"""), '''"x'y\\\"z"'''),
        self.assertEqual(_quoted_if_necessary('foo "bar"'), "'foo \"bar\"'"),
        self.assertEqual(_quoted_if_necessary("""don't break"""), '''"don't break"'''),
        self.assertEqual(_quoted_if_necessary("foo bar"), '"foo bar"')
        self.assertEqual(_quoted_if_necessary("foo"), '"foo"')



"""
Test Cypher conversion system.

Doesn't check validity of output; just matchiness to test strings.
"""

import unittest
import dotmotif
from dotmotif.executors import Neo4jExecutor


_DEMO_G_MIN = """
A -> B
C -> A
B !- A
C !> D
"""
_DEMO_G_MIN_CYPHER = """
MATCH (A:Neuron)-[A_B:SYN]->(B:Neuron)
MATCH (C:Neuron)-[C_A:SYN]->(A:Neuron)
WHERE NOT (B:Neuron)-[:INH]->(A:Neuron) AND NOT (C:Neuron)-[:SYN]->(D:Neuron)
RETURN DISTINCT A,B,C,D
"""

_DEMO_EDGE_ATTR_CYPHER = """
MATCH (A:Neuron)-[A_B:SYN]->(B:Neuron)
MATCH (X:Neuron)-[X_Y:SYN]->(Y:Neuron)
WHERE A_B["weight"] = 4 AND A_B["area"] <= 10 AND X_Y["weight"] = 2
RETURN DISTINCT A,B,X,Y
"""


_DEMO_EDGE_ATTR_CYPHER_2 = """
MATCH (A:Neuron)-[A_B:SYN]->(B:Neuron)
MATCH (X:Neuron)-[X_Y:SYN]->(Y:Neuron)
WHERE A_B["weight"] = 4 AND A_B["area"] <= 10 AND A_B["area"] <= 20 AND X_Y["weight"] = 2
RETURN DISTINCT A,B,X,Y
"""

_DEMO_EDGE_ATTR_CYPHER_2_ENF_INEQ = """
MATCH (A:Neuron)-[A_B:SYN]->(B:Neuron)
MATCH (X:Neuron)-[X_Y:SYN]->(Y:Neuron)
WHERE A_B["weight"] = 4 AND A_B["area"] <= 10 AND A_B["area"] <= 20 AND X_Y["weight"] = 2 AND B<>X AND B<>Y AND A<>X AND A<>B AND A<>Y AND X<>Y
RETURN DISTINCT A,B,X,Y
"""


class TestDotmotif_Cypher(unittest.TestCase):
    def test_cypher(self):
        dm = dotmotif.Motif()
        dm.from_motif(_DEMO_G_MIN)
        self.assertEqual(
            Neo4jExecutor.motif_to_cypher(dm).strip(), _DEMO_G_MIN_CYPHER.strip()
        )


class TestDotmotif_edges_Cypher(unittest.TestCase):
    def test_cypher_edge_attributes(self):
        dm = dotmotif.Motif()
        dm.from_motif(
            """
        A -> B [weight=4, area<=10]
        X -> Y [weight=2]
        """
        )

        self.assertEqual(
            Neo4jExecutor.motif_to_cypher(dm).strip(), _DEMO_EDGE_ATTR_CYPHER.strip()
        )

    def test_cypher_edge_many_attributes(self):
        dm = dotmotif.Motif()
        dm.from_motif(
            """
            A -> B [weight=4, area<=10, area<=20]
            X -> Y [weight=2]
        """
        )

        self.assertEqual(
            Neo4jExecutor.motif_to_cypher(dm).strip(), _DEMO_EDGE_ATTR_CYPHER_2.strip()
        )

    #   TODO      # Issue with arbitrary ordering of inequalities
    # def test_cypher_edge_many_attributes_and_enforce_inequality(self):
    #     dm = dotmotif.Motif(enforce_inequality=True)
    #     dm.from_motif(
    #         """
    #         A -> B [weight=4, area<=10, area<=20]
    #         X -> Y [weight=2]
    #     """
    #     )

    #     self.assertEqual(
    #         Neo4jExecutor.motif_to_cypher(dm).strip(),
    #         _DEMO_EDGE_ATTR_CYPHER_2_ENF_INEQ.strip(),
    #     )


class TestDotmotif_nodes_Cypher(unittest.TestCase):
    def test_cypher_node_attributes(self):
        dm = dotmotif.Motif()
        dm.from_motif(
            """
        A -> B
        A.size = "big"
        """
        )

        self.assertEqual(
            Neo4jExecutor.motif_to_cypher(dm).strip(),
            """MATCH (A:Neuron)-[A_B:SYN]->(B:Neuron)\nWHERE A["size"] = "big"\nRETURN DISTINCT A,B""".strip(),
        )

    def test_cypher_node_many_attributes(self):
        dm = dotmotif.Motif()
        dm.from_motif(
            """
        A -> B
        A.size = "big"
        A.type = "funny"
        """
        )

        self.assertEqual(
            Neo4jExecutor.motif_to_cypher(dm).strip(),
            """MATCH (A:Neuron)-[A_B:SYN]->(B:Neuron)\nWHERE A["size"] = "big" AND A["type"] = "funny"\nRETURN DISTINCT A,B""".strip(),
        )

    def test_cypher_node_same_node_many_attributes(self):
        dm = dotmotif.Motif()
        dm.from_motif(
            """
        A -> B
        A.personality != "exciting"
        A.personality != "funny"
        """
        )

        self.assertEqual(
            Neo4jExecutor.motif_to_cypher(dm).strip(),
            """MATCH (A:Neuron)-[A_B:SYN]->(B:Neuron)\nWHERE A["personality"] <> "exciting" AND A["personality"] <> "funny"\nRETURN DISTINCT A,B""".strip(),
        )

    def test_cypher_node_many_node_attributes(self):
        dm = dotmotif.Motif()
        dm.from_motif(
            """
        A -> B
        A.area <= 10
        B.area <= 10
        """
        )

        self.assertEqual(
            Neo4jExecutor.motif_to_cypher(dm).strip(),
            """MATCH (A:Neuron)-[A_B:SYN]->(B:Neuron)\nWHERE A["area"] <= 10 AND B["area"] <= 10\nRETURN DISTINCT A,B""".strip(),
        )

    def test_cypher_negative_edge_and_inequality(self):
        dm = dotmotif.Motif(enforce_inequality=True)
        dm.from_motif(
            """
        A -> B
        A -> C
        B !> C
        """
        )

        self.assertEqual(
            Neo4jExecutor.motif_to_cypher(dm).strip(),
            (
                "MATCH (A:Neuron)-[A_B:SYN]->(B:Neuron)\n"
                "MATCH (A:Neuron)-[A_C:SYN]->(C:Neuron)\n"
                "WHERE NOT (B:Neuron)-[:SYN]->(C:Neuron) AND A<>B AND A<>C AND B<>C\n"
                "RETURN DISTINCT A,B,C"
            ).strip(),
        )


class TestDotmotif_nodes_edges_Cypher(unittest.TestCase):
    def test_cypher_node_and_edge_attributes(self):
        dm = dotmotif.Motif()
        dm.from_motif(
            """
        A -> B [area != 10]
        A.area <= 10
        B.area <= 10
        """
        )

        self.assertEqual(
            Neo4jExecutor.motif_to_cypher(dm).strip(),
            (
                "MATCH (A:Neuron)-[A_B:SYN]->(B:Neuron)\n"
                'WHERE A["area"] <= 10 AND B["area"] <= 10 AND A_B["area"] <> 10\n'
                "RETURN DISTINCT A,B"
            ).strip(),
        )


class TestDynamicNodeConstraints(unittest.TestCase):
    def test_dynamic_constraints_in_cypher(self):
        dm = dotmotif.Motif(enforce_inequality=True)
        dm.from_motif(
            """
        A -> B
        A.radius >= B.radius
        A.zorf != B.zorf
        """
        )
        self.assertIn(
            """WHERE A["radius"] >= B["radius"] AND A["zorf"] <> B["zorf"]""",
            Neo4jExecutor.motif_to_cypher(dm).strip(),
        )


class BugReports(unittest.TestCase):
    def test_fix_where_clause__github_35(self):
        dm = dotmotif.Motif(enforce_inequality=True)
        dm.from_motif(
            """
        A -> B
        """
        )
        self.assertIn("WHERE", Neo4jExecutor.motif_to_cypher(dm).strip())



import unittest
import dotmotif


class TestDotmotifIllegal(unittest.TestCase):
    """
    Test illegal dotmotif operations.
    """

    def test_disagreeing_edges(self):
        dm = dotmotif.Motif(
            validators=[dotmotif.validators.DisagreeingEdgesValidator()]
        )
        with self.assertRaises(Exception):
            dm.from_motif("A -> B\nA !> B")

    def test_disagreeing_edges_ignored(self):
        dm = dotmotif.Motif(validators=[])
        dm.from_motif("A -+ B\nA !> B")



import unittest
import dotmotif


_THREE_CYCLE = """A -> B\nB -> C\nC -> A\n"""
_THREE_CYCLE_NEG = """A !> B\nB !> C\nC !> A\n"""
_THREE_CYCLE_INH = """A -| B\nB -| C\nC -| A\n"""
_THREE_CYCLE_NEG_INH = """A !| B\nB !| C\nC !| A\n"""
_ABC_TO_D = """\nA -> D\nB -> D\nC -> D\n"""

_THREE_CYCLE_CSV = """\nA,B\nB,C\nC,A\n"""
_THREE_CYCLE_NEG_CSV = """\nA,B\nB,C\nC,A\n"""


class TestDotmotif_DM(unittest.TestCase):
    def test_sanity(self):
        self.assertEqual(1, 1)

    def test_dm_parser(self):
        dm = dotmotif.Motif(_THREE_CYCLE)
        self.assertEqual(len(dm._g.edges()), 3)
        self.assertEqual(len(dm._g.nodes()), 3)

    def test_dm_parser_actions(self):
        dm = dotmotif.Motif(_THREE_CYCLE)
        self.assertEqual([e[2]["action"] for e in dm._g.edges(data=True)], ["SYN"] * 3)

        dm = dotmotif.Motif(_THREE_CYCLE_INH)
        self.assertEqual([e[2]["action"] for e in dm._g.edges(data=True)], ["INH"] * 3)

    def test_dm_parser_edge_exists(self):
        dm = dotmotif.Motif(_THREE_CYCLE)
        self.assertEqual([e[2]["exists"] for e in dm._g.edges(data=True)], [True] * 3)

        dm = dotmotif.Motif(_THREE_CYCLE_NEG)
        self.assertEqual([e[2]["exists"] for e in dm._g.edges(data=True)], [False] * 3)

        dm = dotmotif.Motif(_THREE_CYCLE_NEG_INH)
        self.assertEqual([e[2]["exists"] for e in dm._g.edges(data=True)], [False] * 3)



# Standard installs:
import abc
import json
import os
import numbers

# Non-standard installs:
import dask.dataframe as dd
import pandas as pd

# Types only:
from typing import List
import networkx as nx

from .. import utils


class NetworkXConverter(abc.ABC):
    """
    An abstract base class for import to the NetworkX format.

    """

    pass

    def to_graph(self) -> nx.Graph:
        ...


class EdgelistConverter(NetworkXConverter):
    """
    Convert an edgelist dataframe or CSV to a graph.

    """

    def __init__(
        self,
        filepath_or_dataframe: str,
        u_id_column: str,
        v_id_column: str,
        directed: bool = True,
        file_reader_kwargs: dict = None,
    ):
        if isinstance(filepath_or_dataframe, pd.DataFrame):
            data = filepath_or_dataframe
        else:
            data = pd.read_table(
                filepath_or_dataframe,
                dtype={u_id_column: str, v_id_column: str},
                **(file_reader_kwargs or {"sep": ","}),
            )
        self._graph = nx.DiGraph() if directed else nx.Graph()
        for i, row in data.iterrows():
            self._graph.add_edge(
                str(row[u_id_column]), str(row[v_id_column]), **dict(row)
            )

    def to_graph(self):
        return self._graph


CSVEdgelistConverter = EdgelistConverter


class Ingester:
    """
    Base ingester class
    """

    def ingest(self) -> List[str]:
        """
        Ingest the data and output a list of files where the data were saved.
        """
        ...


class FileIngester(Ingester):
    """
    Base class for ingester from a file on disk (or file-like).

    Abstract class, should not be used directly.
    """

    def __init__(self, path: str, export_dir: str) -> None:
        """
        Store the path and export directory of this ingester.

        All ingesters must have an input file(s) and an output directory which
        can be used by the dotmotif class for input to, say, a Neo4j docker
        container, among other platforms.
        """
        self.path = path
        self.export_dir = export_dir.rstrip("/") + "/"
        if not os.path.isdir(self.export_dir):
            os.makedirs(self.export_dir)


class NetworkXIngester:
    """
    An ingester that converts an in-memory networkx graph into two files:

    One file contains the node IDs and their attrs, and the other contains
    the edges and the edge attrs. Each column must contain a type annotation
    in order for Neo4j to be able to interpret the attributes as variables.
    """

    def __init__(self, graph: nx.Graph, export_dir: str) -> None:
        """
        Create a new ingester from networkx format.

        Arguments:
            graph (nx.Graph): The in-memory graph to ingest
            export_dir (str): The path to the CSV directory to use

        """
        self.export_dir = export_dir.rstrip("/") + "/"
        self.graph = graph

    def ingest(self) -> List[str]:
        def _pd_cypher_dtype_map(dtype):
            _valid_cypher_dtypes = [
                # https://neo4j.com/docs/operations-manual/current/tools/import/file-header-format/
                "int",
                "long",
                "float",
                "double",
                "boolean",
                "byte",
                "short",
                # "char", "string",
                # "point", "date", "localtime", "time", "localdatetime", "datetime", "duration"
            ]
            # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.api.types.infer_dtype.html
            return {
                "object": "string",
                "string": "string",
                "bytes": "string",
                "mixed-integer-float": "float",
                "floating": "float",
                "integer": "int",
                "mixed-integer": "int",
            }.get(pd.api.types.infer_dtype(dtype, skipna=True), "string")

        # First make sure that none of the IDs are numbers:
        for n in self.graph.nodes():
            if isinstance(n, (int, float)):
                raise ValueError(
                    "DotMotif does not support haystack graphs with numerical IDs. "
                    + "Not all executors can operate on numerical-ID graphs, and so "
                    + "they are disallowed. "
                    + "Learn more at https://github.com/aplbrain/dotmotif/issues/26"
                )

        # Now create a node-attribute dataframe:
        nodes_df = pd.DataFrame(
            {"neuronId:ID(Neuron)": u, **attrs}
            for u, attrs in self.graph.nodes(data=True)
        )

        # And edges:
        edges_df = pd.DataFrame(
            {":START_ID(Neuron)": u, ":END_ID(Neuron)": v, **attrs}
            for u, v, attrs in self.graph.edges(data=True)
        )

        edges_df.columns = [
            f"{c}"
            + (f":{_pd_cypher_dtype_map(edges_df[c])}" if "(Neuron)" not in c else "")
            for c in edges_df.columns
        ]

        nodes_df.columns = [
            f"{c}"
            + (f":{_pd_cypher_dtype_map(nodes_df[c])}" if "(Neuron)" not in c else "")
            for c in nodes_df.columns
        ]

        # Export the files:
        os.makedirs(f"{self.export_dir}", exist_ok=True)

        nodes_df.to_csv(f"{self.export_dir}/export-neurons-0.csv")
        edges_df.to_csv(f"{self.export_dir}/export-synapses-zdata.csv")

        return [
            f"{self.export_dir}/export-neurons-0.csv",
            f"{self.export_dir}/export-synapses-zdata.csv",
        ]


class _deprecated_Type2FileIngester(FileIngester):
    """
    An Ingester that reads data in the Type 2 CSV file format.

    The data are an edge-list, with extra columns that are currently
    ignored after ingest.
    """

    def __init__(self, path: str, export_dir: str) -> None:
        """
        Create a new ingester.

        .
        """
        super().__init__(path, export_dir)

    def ingest(self) -> List[str]:
        """
        Ingest using dask and pandas.

        .
        """
        df = dd.read_csv(self.path).dropna()
        export_df = df.copy()
        export_df[":START_ID(Neuron)"] = df["presyn_segid"].astype("int")
        export_df[":END_ID(Neuron)"] = df["postsyn_segid"].astype("int")
        for col in df.columns:
            del export_df[col]

        node_names = (
            (export_df[":START_ID(Neuron)"].append(export_df[":END_ID(Neuron)"]))
            .unique()
            .dropna()
        )

        node_fnames = node_names.to_csv(
            self.export_dir + "export-neurons-*.csv",
            index=False,
            header=["neuronId:ID(Neuron)"],
        )

        # This is absurd, but neo4j can't tolerate file headers in every CSV,
        # and dask can't NOT.
        # So We print off a header file first.
        headerpath = self.export_dir + "export-synapses-header.csv"
        with open(headerpath, "w") as headerfile:
            headerfile.write(":START_ID(Neuron),:END_ID(Neuron)")

        edge_fnames = export_df.to_csv(
            self.export_dir + "export-synapses-zdata-*.csv", index=False, header=False
        )

        return node_fnames + [headerpath] + edge_fnames


class _deprecated_Type1FileIngester(FileIngester):
    def __init__(self, path: str, export_dir: str) -> None:
        """
        Create a new ingester.
        """
        super().__init__(path, export_dir)

    def ingest(self) -> List[str]:
        """
        Ingest using dask and pandas.
        """
        data = json.load(open(self.path, "r"))
        export_df = pd.DataFrame(
            {":START_ID(Neuron)": data["neuron_1"], ":END_ID(Neuron)": data["neuron_2"]}
        )

        node_names = (
            export_df[":START_ID(Neuron)"].append(export_df[":END_ID(Neuron)"])
        ).unique()

        node_names_dd = dd.from_pandas(
            pd.DataFrame({"neuronId:ID(Neuron)": node_names}), npartitions=1
        )

        node_fnames = node_names_dd.to_csv(
            self.export_dir + "export-neurons-*.csv",
            index=False,
            header=["neuronId:ID(Neuron)"],
        )

        # This is absurd, but neo4j can't tolerate file headers in every CSV,
        # and dask can't NOT.
        # So We print off a header file first.
        export_df_dd = dd.from_pandas(export_df, npartitions=1)
        headerpath = self.export_dir + "export-synapses-header.csv"
        with open(headerpath, "w") as headerfile:
            headerfile.write(":START_ID(Neuron),:END_ID(Neuron)")

        edge_fnames = export_df_dd.to_csv(
            self.export_dir + "export-synapses-zdata-*.csv", index=False, header=False
        )

        return node_fnames + [headerpath] + edge_fnames



"""
Copyright 2021 The Johns Hopkins University Applied Physics Laboratory.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.`
"""

from typing import TYPE_CHECKING, List, Tuple
import copy
import networkx as nx

from .Executor import Executor

if TYPE_CHECKING:
    from .. import dotmotif

_OPERATORS = {
    "=": lambda x, y: x == y,
    "==": lambda x, y: x == y,
    ">=": lambda x, y: x >= y,
    "<=": lambda x, y: x <= y,
    "<": lambda x, y: x < y,
    ">": lambda x, y: x > y,
    "!=": lambda x, y: x != y,
    "in": lambda x, y: x in y,
    "contains": lambda x, y: y in x,
    "!in": lambda x, y: x not in y,
    "!contains": lambda x, y: y not in x,
}


def _edge_satisfies_constraints(edge_attributes: dict, constraints: dict) -> bool:
    """
    Check if a single edge satisfies the constraints.
    """

    for key, clist in constraints.items():
        for operator, values in clist.items():
            for value in values:
                keyvalue_or_none = edge_attributes.get(key, None)
                try:
                    operator_success = _OPERATORS[operator](keyvalue_or_none, value)
                except TypeError:
                    # If you encounter a type error, that means the comparison
                    # could not possibly succeed,
                    # # TODO: unless you tried a comparison
                    # against an undefined value (i.e. VALUE != undefined)
                    return False
                if not operator_success:
                    # Fail fast, if any edge attributes fail the test
                    return False
    return True


def _edge_satisfies_many_constraints_for_muligraph_any_edges(edge_attributes: dict, constraints: dict) -> List[Tuple[str, str, str]]:
    """
    Returns a subset of constraints that this edge matches, in the form (key, op, val).
    """
    matched_constraints = []
    for key, clist in constraints.items():
        for operator, values in clist.items():
            for value in values:
                keyvalue_or_none = edge_attributes.get(key, None)
                try:
                    operator_success = _OPERATORS[operator](keyvalue_or_none, value)
                except TypeError:
                    # If you encounter a type error, that means the comparison
                    # could not possibly succeed
                    # # TODO: unless you tried a comparison
                    # against an undefined value (i.e. VALUE != undefined)
                    operator_success = False
                if operator_success:
                    matched_constraints.append((key, operator, value))
    return matched_constraints

def _node_satisfies_constraints(node_attributes: dict, constraints: dict) -> bool:
    """
    Check if a single node satisfies the constraints.

    """
    for key, clist in constraints.items():
        for operator, values in clist.items():
            for value in values:
                if not _OPERATORS[operator](node_attributes.get(key, None), value):
                    # Fail fast, if any node attributes fail the test
                    return False
    return True


class NetworkXExecutor(Executor):
    """
    A query executor that runs inside RAM.

    Uses NetworkX's built-in (VF2) subgraph isomorphism algo. Good for very
    small graphs, since this won't scale particularly well.
    """

    def __init__(self, **kwargs) -> None:
        """
        Create a new NetworkXExecutor.

        Arguments:
            graph (networkx.Graph)
            multigraph_edge_match (str: 'any'): A string ('any' or 'all') that
                determines how to match edges between nodes in the graph. If
                'any', then any edge between nodes can match the constraints
                to satisfy the motif. If 'all', then all edges between nodes
                must match the constraints to satisfy the motif.

        Returns:
            None

        """
        if "graph" in kwargs:
            self.graph = kwargs.get("graph")
        else:
            raise ValueError(
                "You must pass a graph to the NetworkXExecutor constructor."
            )

        # Allow the user to set whether ALL edges should match when considering
        # a multigraph or if ANY edges should be considered when matching.
        # We only do this if the graph is a multigraph.
        self._host_is_multigraph = False
        if self.graph.is_multigraph():
            self._host_is_multigraph = True
            self._multigraph_edge_match = kwargs.get("multigraph_edge_match", "any")
            assert self._multigraph_edge_match in ("all", "any"), "_multigraph_edge_match must be one of 'all' or 'any'."

    def _validate_node_constraints(
        self, node_isomorphism_map: dict, graph: nx.DiGraph, constraints: dict
    ) -> bool:
        """
        Validate nodes against their isomorphism's constraints in the motif.

        Arguments:
            ...

        Returns:
            bool
        """
        for motif_U, constraint_list in constraints.items():
            graph_u = node_isomorphism_map[motif_U]

            if not _node_satisfies_constraints(graph.nodes[graph_u], constraint_list):
                return False
        return True

    def _validate_dynamic_node_constraints(
        self, node_isomorphism_map: dict, graph: nx.DiGraph, constraints: dict
    ) -> bool:
        """
        Validate a graph against its dynamic node constraints.

        Dynamic node constraints are constraints that compare two attributes
        from the graph (rather than one key/val from the graph and a static
        value that is known a priori).

        Arguments:
            ...

        Returns:
            bool
        """
        # `constraints` is of the form:
        # { thisNodeId: { thisKey: { operator: [ ( thatNodeId, thatKey )]}}}
        for motif_U, constraint_list in constraints.items():
            this_node = node_isomorphism_map[motif_U]
            for this_key, operators in constraint_list.items():
                for operator, that_node_list in operators.items():
                    for (that_node_V, that_key) in that_node_list:
                        that_node = node_isomorphism_map[that_node_V]
                        if this_key not in graph.nodes[this_node]:
                            return False
                        if that_key not in graph.nodes[that_node]:
                            return False
                        if not _OPERATORS[operator](
                            graph.nodes[this_node][this_key],
                            graph.nodes[that_node][that_key],
                        ):
                            return False
        return True

    def _validate_edge_constraints(
        self, node_isomorphism_map: dict, graph: nx.DiGraph, constraints: dict
    ):
        """
        Validate all edge constraints on a subgraph.

        Arguments:
            node_isomorphism_map (dict[nodename:str, nodeID:str]): A mapping of
                node names to node IDs (where name comes from the motif and the
                ID comes from the haystack graph).
            graph (nx.DiGraph): The haystack graph
            constraints (dict[(motif_u, motif_v), dict[operator, value]]): Map
                of constraints on the MOTIF node names.

        Returns:
            bool: If the isomorphism satisfies the edge constraints

        For example, if constraints =
        {
            ("A", "B"): {"weight": { "==": 10 }}
        }

        And node_isomorphism_map =
        {
            "A": "x",
            "B": "y",
        }

        And haystack contains the edge (x, y) with attribute weight=10, then
        this function will return True.

        """
        for (motif_U, motif_V), constraint_list in constraints.items():
            # Get graph nodes (from this isomorphism)
            graph_u = node_isomorphism_map[motif_U]
            graph_v = node_isomorphism_map[motif_V]

            # Check edge in graph for constraints
            edge_attrs = graph.get_edge_data(graph_u, graph_v)

            if not _edge_satisfies_constraints(edge_attrs, constraint_list):
                # Fail fast
                return False
        return True

    def _validate_multigraph_all_edge_constraints(
        self, node_isomorphism_map: dict, graph: nx.DiGraph, constraints: dict
    ):
        """
        Reuses logic from the simple _validate_edge_constraints case.

        Sole modification is that in the multigraph case, ALL edges must match
        for an edge to be considered valid. If ANY of the edges between two
        nodes mismatch the constraints, the mapping fails.

        """
        for (motif_U, motif_V), constraint_list in constraints.items():
            # Get graph nodes (from this isomorphism)
            graph_u = node_isomorphism_map[motif_U]
            graph_v = node_isomorphism_map[motif_V]

            # Check each edge in graph for constraints
            for _, _, edge_attrs in graph.edges((graph_u, graph_v), data=True):
                if not _edge_satisfies_constraints(edge_attrs, constraint_list):
                    # Fail fast
                    return False
        return True

    def _validate_multigraph_any_edge_constraints(
        self, node_isomorphism_map: dict, graph: nx.DiGraph, constraints: dict
    ):
        """
        Reuses logic from the simple _validate_edge_constraints case.

        Sole modification is that in the multigraph case, ANY edge can match
        for an edge to be considered valid. If ANY of the edges between two
        nodes match the constraints, the mapping succeeds.

        """
        for (motif_U, motif_V), constraint_list in constraints.items():
            # Get graph nodes (from this isomorphism)
            graph_u = node_isomorphism_map[motif_U]
            graph_v = node_isomorphism_map[motif_V]

            # check each edge in the graph for the constraints.
            # if you find an edge that matches, REMOVE that constraint from
            # the list and continue checking.
            # if you get to the end of the list of edges and there are any
            # constrains left, the mapping fails.

            # Check each edge in graph for constraints
            constraint_list_copy = copy.deepcopy(constraint_list)
            for _, _, edge_attrs in graph.edges((graph_u, graph_v), data=True):
                matched_constraints = _edge_satisfies_many_constraints_for_muligraph_any_edges(edge_attrs, constraint_list_copy)
                if matched_constraints:
                    # Remove matched constraints from the list
                    for constraint in matched_constraints:
                        (key, operator, value) = constraint
                        # remove `value` from the list of [key][operator].
                        # if the list is empty, remove the entire key.
                        constraint_list_copy[key][operator].remove(value)
                        if len(constraint_list_copy[key][operator]) == 0:
                            del constraint_list_copy[key][operator]
                        if not constraint_list_copy[key]:
                            del constraint_list_copy[key]

            # if there are any constraints left over, the mapping failed
            if len(constraint_list_copy) > 0:
                return False

        return True


    def count(self, motif: "dotmotif.Motif", limit: int = None):
        """
        Count the occurrences of a motif in a graph.

        See NetworkXExecutor#find for more documentation.
        """
        return len(self.find(motif, limit))

    def find(self, motif: "dotmotif.Motif", limit: int = None):
        """
        Find a motif in a larger graph.

        Arguments:
            motif (dotmotif.Motif)

        """
        # TODO: Can add constraints on iso node assignment. If we do this a
        # little smarter, can save a lot of post-processing time-complexity.

        # We need to first remove "negative" nodes from the motif, and then
        # filter them out later on. Though this reduces the speed of the graph-
        # matching, NetworkX does not seem to support this out of the box.
        # TODO: Confirm that networkx does not support this out of the box.

        if motif.ignore_direction or not self.graph.is_directed:
            graph_constructor = nx.Graph
            graph_matcher = nx.algorithms.isomorphism.GraphMatcher
        else:
            graph_constructor = nx.DiGraph
            graph_matcher = nx.algorithms.isomorphism.DiGraphMatcher

        only_positive_edges_motif = graph_constructor()
        must_not_exist_edges = []
        for u, v, attrs in motif.to_nx().edges(data=True):
            if attrs["exists"] is True:
                only_positive_edges_motif.add_edge(u, v, **attrs)
            elif attrs["exists"] is False:
                # Collect a list of neg-edges to check for again in a moment
                must_not_exist_edges.append((u, v))
        gm = graph_matcher(self.graph, only_positive_edges_motif)

        def _doesnt_have_any_of_motifs_negative_edges(mapping):
            for u, v in must_not_exist_edges:
                if self.graph.has_edge(mapping[u], mapping[v]):
                    return False
            return True

        unfiltered_results = [
            # Here, `mapping` has keys of self.graph node IDs and values of
            # motif node names. We need the reverse for pretty much everything
            # we do from here out, so we reverse the pairs.
            {v: k for k, v in mapping.items()}
            # TODO: Use isomorphism here if requested
            for mapping in gm.subgraph_monomorphisms_iter()
        ]

        # Now, filter out those that have edges they should not:
        results = [
            mapping
            for mapping in unfiltered_results
            if _doesnt_have_any_of_motifs_negative_edges(mapping)
        ]

        _edge_constraint_validator = (
            self._validate_edge_constraints if not self._host_is_multigraph else (
                self._validate_multigraph_all_edge_constraints
                if self._multigraph_edge_match == "all"
                else self._validate_multigraph_any_edge_constraints
            )
        )
        # Now, filter on attributes:
        res = [
            r
            for r in results
            if (
                _edge_constraint_validator(
                    r, self.graph, motif.list_edge_constraints()
                )
                and self._validate_node_constraints(
                    r, self.graph, motif.list_node_constraints()
                )
                and self._validate_dynamic_node_constraints(
                    r, self.graph, motif.list_dynamic_node_constraints()
                )
                # by default, networkx returns the automorphism that is left-
                # sorted, so this comparison is _opposite_ the check that we
                # use in the other executors. In other words, we usually check
                # that A >= B; here we check A <= B.
                and (
                    (not motif.exclude_automorphisms)
                    or all(r[a] <= r[b] for (a, b) in motif.list_automorphisms())
                )
            )
        ]
        return res



#!/usr/bin/env python3

"""
Utilities for dotmotif.

Requires nx>=2.0
"""
import networkx as nx


def untype_string(string):
    try:
        return eval(string)
    except:
        try:
            if int(string) == float(string):
                return int(string)
            return float(string)
        except:
            return str(string)


def draw_motif(dm, negative_edge_color: str = "r", pos=None):
    """
    Draw a dotmotif motif object.

    Arguments:
        dm: dotmotif.Motif
        negative_edge_color (str: r): Color used to represent negative edges
        pos (dict: None): The position to use. If unset, uses nx.spring_layout

    Returns:
        None

    """
    if pos is None:
        pos = nx.spring_layout(dm._g)
    exc_edges = list(
        filter(lambda e: e[2].get("action", "") == "SYN", dm._g.edges(data=True))
    )
    inh_edges = list(
        filter(lambda e: e[2].get("action", "") != "SYN", dm._g.edges(data=True))
    )
    nx.draw_networkx_edges(
        dm._g,
        pos=pos,
        edgelist=exc_edges,
        edge_color=[
            "k" if e["exists"] else negative_edge_color for u, v, e in exc_edges
        ],
    )
    nx.draw_networkx_edges(
        dm._g,
        pos=pos,
        edgelist=inh_edges,
        edge_color=[
            "k" if e["exists"] else negative_edge_color for u, v, e in inh_edges
        ],
        arrowstyle="-[",
    )
    nx.draw_networkx_nodes(dm._g, pos=pos, c="b")
    nx.draw_networkx_labels(dm._g, pos=pos)



#!/usr/bin/env python3
"""
Copyright 2022 The Johns Hopkins University Applied Physics Laboratory.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Union, IO
import copy
import pickle
import warnings

import networkx as nx
from networkx.algorithms import isomorphism

from .parsers.v2 import ParserV2
from .validators import DisagreeingEdgesValidator

from .executors.NetworkXExecutor import NetworkXExecutor
from .executors.GrandIsoExecutor import GrandIsoExecutor
from .executors.Neo4jExecutor import Neo4jExecutor

__version__ = "0.11.0"

DEFAULT_MOTIF_PARSER = ParserV2


class MotifError(ValueError):
    pass


class Motif:
    """
    Container class for dotmotif operations.

    See __init__ documentation for more details.
    """

    def __init__(self, input_motif: str = None, **kwargs):
        """
        Create a new dotmotif object.

        Arguments:
            input_motif (str: None): Optionally, a DotMotif DSL defined motif,
                or a path to a .motif file that contains a motif.
            ignore_direction (bool: False): Whether to disregard direction when
                generating the database query
            limit (int: None): A limit (if any) to impose on the query results
            enforce_inequality (bool: False): Whether to enforce inequality; in
                other words, whether two nodes should be permitted to be aliases
                for the same node. For example, A>B>C; if A!=C, then set to True
            pretty_print (bool: True)
            parser (dotmotif.parsers.Parser: DEFAULT_MOTIF_PARSER): The parser
                to use to parse the document. Defaults to the v2 parser.
            exclude_automorphisms (bool: False): Whether to exclude automorphism
                variants of the motif when returning results.
            validators (List[Validator]): A list of dotmotif.Validators to use
                when verifying the motif for correctness and executability.

        """
        self.ignore_direction = kwargs.get("ignore_direction", False)
        self.limit = kwargs.get("limit", None)
        self.enforce_inequality = kwargs.get("enforce_inequality", False)
        self.pretty_print = kwargs.get("pretty_print", True)
        self.parser = kwargs.get("parser", DEFAULT_MOTIF_PARSER)
        self.exclude_automorphisms = kwargs.get("exclude_automorphisms", False)
        self.validators = kwargs.get("validators", [DisagreeingEdgesValidator()])
        self._g = nx.MultiDiGraph()

        self._edge_constraints = {}
        self._node_constraints = {}
        self._dynamic_edge_constraints = {}
        self._dynamic_node_constraints = {}
        self._automorphisms = []

        if input_motif:
            self.from_motif(input_motif)

    def from_motif(self, cmd: str):
        """
        Ingest a dotmotif-format string.

        Arguments:
            cmd (str): A string in dotmotif form, or a .dm filename on disk

        Returns:
            A pointer to this dotmotif object, for chaining

        """
        if len(cmd.split("\n")) == 1:
            try:
                cmd = open(cmd, "r").read()
            except FileNotFoundError:
                pass

        result = self.parser(validators=self.validators).parse(cmd)
        if isinstance(result, tuple):
            (
                self._g,
                self._edge_constraints,
                self._node_constraints,
                self._dynamic_edge_constraints,
                self._dynamic_node_constraints,
                self._automorphisms,
            ) = result
        else:
            # For backwards compatibility with parser v1
            self._g = result

        return self

    def from_nx(self, graph: nx.DiGraph) -> "Motif":
        """
        Ingest directly from a graph.

        Arguments:
            graph (nx.DiGraph): The graph to import

        Returns:
            None

        """

        warnings.warn(
            "The dotmotif#from_nx call is deprecated as of v0.4.3. "
            "For more information, please read here: "
            "https://github.com/aplbrain/dotmotif/issues/43",
            DeprecationWarning,
        )
        self._g = copy.deepcopy(graph)
        for u, v, edge_attrs in self._g.edges(data=True):
            if "exists" not in edge_attrs:
                self._g.edges[u, v]["exists"] = True
        return self

    def to_nx(self) -> nx.DiGraph:
        """
        Output a networkx graph describing the motif.

        Returns:
            networkx.DiGraph

        """
        return self._g

    def list_edge_constraints(self):
        return self._edge_constraints

    def list_dynamic_edge_constraints(self):
        return self._dynamic_edge_constraints

    def list_node_constraints(self):
        return self._node_constraints

    def list_dynamic_node_constraints(self):
        return self._dynamic_node_constraints

    def list_automorphisms(self):
        if not self.exclude_automorphisms:
            return self._automorphisms

        g = self.to_nx()
        res = isomorphism.GraphMatcher(g, g).subgraph_isomorphisms_iter()
        autos = set()
        for auto in res:
            for k, v in auto.items():
                if k != v:
                    autos.add(tuple(sorted([k, v])))
        return list(autos)

    def save(self, fname: Union[str, IO[bytes]]) -> Union[str, IO[bytes]]:
        """
        Save the motif to a file on disk.

        Arguments:
            fname (str): A path on disk for IO

        Returns:
            Pointer to File-like.

        """
        if isinstance(fname, str):
            f = open(fname, "wb")
        else:
            f = fname
        pickle.dump(self, f)
        return fname

    @staticmethod
    def load(fname: Union[str, IO[bytes]]) -> "Motif":
        """
        Load the motif from a file on disk.

        Arguments:
            fname (str): A path on disk for IO

        Returns:
            Pointer to File-like.

        """
        if isinstance(fname, str):
            f = open(fname, "rb")
        else:
            f = fname
        result = pickle.load(f)
        f.close()
        return result



import os

HOST = "neuprint.janelia.org"
DATASET = "hemibrain:v1.1"
TOKEN = os.getenv("NEUPRINT_TOKEN")

import unittest

from .. import Motif
from .NeuPrintExecutor import NeuPrintExecutor

if TOKEN:

    class TestNeuPrintConnection(unittest.TestCase):
        def test_can_get_version(self):
            E = NeuPrintExecutor(HOST, DATASET, TOKEN)
            self.assertEqual(E.client.fetch_version(), "0.1.0")

    class TestNeuPrintKnownMotifs(unittest.TestCase):
        def test_known_edge(self):
            motif = Motif().from_motif(
                """
                input -> output
                input.type == "KCab-p"
                """
            )
            E = NeuPrintExecutor(HOST, DATASET, TOKEN)

            self.assertTrue("ConnectsTo" in E.motif_to_cypher(motif))
            self.assertEqual(len(E.find(motif, limit=5)), 5)

        def test_bigger_motif(self):
            motif = Motif().from_motif(
                """
                diedge(a, b) {
                    a -> b
                    b !> a
                    a.type == "KCab-p"
                }
                diedge(A, B)
                diedge(B, C)
                diedge(C, D)

                A !> C
                C !> A
                D !> B
                B !> D
                """
            )
            E = NeuPrintExecutor(HOST, DATASET, TOKEN)

            self.assertTrue("ConnectsTo" in E.motif_to_cypher(motif))
            print(E.motif_to_cypher(motif))
            self.assertEqual(E.count(motif, limit=50), 35)



#!/usr/bin/env python3
import abc

import networkx as nx


class Validator(abc.ABC):
    @abc.abstractmethod
    def validate(self) -> bool:
        ...


class NeuronConnectomeValidator(Validator):
    class NeuronConnectomeValidatorError(Exception):
        pass

    def __init__(self, g: nx.Graph) -> None:
        self.g = g

    def validate(self) -> bool:
        """
        Return True if valid.

        Raises NeuronConnectomeValidatorError if invalid.
        """
        return True


class DisagreeingEdgesValidatorError(Exception):
    pass


class DisagreeingEdgesValidator(Validator):
    def __init__(self) -> None:
        pass

    def validate(self, g: nx.Graph, u, v, type=None, exists: bool = True) -> bool:
        """
        Return True if valid.

        Raises DisagreeingEdgesValidatorError if invalid.
        """
        edges = g.edges(u, data=True)
        for u_, v_, attrs in edges:
            if u_ == u and v_ == v and exists != attrs["exists"]:
                raise DisagreeingEdgesValidatorError(
                    f"Trying to add <{u}-{v} exists={exists}> but "
                    f"<{u_}-{v_} exists={attrs['exists']}> already in motif."
                )
        return True



#!/usr/bin/env python3
from typing import List
import warnings

import networkx as nx

from .. import Parser
from ...validators import Validator


class ParserDMv1(Parser):
    """
    Version 1 of the DotMotif DSL parser.

    Deprecated as of 0.4.3 (Sept 2019).
    """

    _LOOKUP = {"INHIBITS": "INH", "EXCITES": "EXC", "SYNAPSES": "SYN"}

    ACTIONS = {
        ">": "SYNAPSES",
        "~": "INHIBITS",
        "|": "INHIBITS",
        "+": "EXCITES",
        "?": "SYNAPSES",
        "-": "SYNAPSES",
    }

    def __init__(self, validators: List[Validator]) -> None:
        """
        Create a new Parser using the V1 service.

        Now deprecated.
        """
        warnings.warn(
            "Version 1 of the DotMotif DSL parser is deprecated as of v0.4.3. "
            "Please update your code to use `dotmotif.parsers.v2.ParserV2`, "
            "or you can point to `dotmotif.DEFAULT_MOTIF_PARSER` "
            "(which will always point to the latest DSL parser available.)",
            DeprecationWarning,
        )
        self.validators = validators

    def parse(self, dm: str) -> nx.MultiDiGraph:
        """
        Parse the DotMotif using the old V1 DSL parser.

        Now deprecated.
        """
        G = nx.MultiDiGraph()
        for line in dm.split("\n"):
            res = self._parse_dm_line(line.strip())
            if res:
                u, v, action, exists = res
                for val in self.validators:
                    val.validate(G, u, v, action, exists)
                G.add_edge(u, v, exists=exists, action=action)
        return G

    def _parse_dm_line(self, _line: str):
        _line = _line.strip()
        # Tokenize:
        if len(_line) and _line.startswith("#"):
            return None
        line = [t for t in _line.split() if len(t)]
        if len(line) == 0:
            return None

        # Format should be [NEURON_ID, ACTION, NEURON_ID]
        try:
            u, action, v = line
        except ValueError:
            raise ValueError(
                "Line must be of the form [NEURON_ID, ACTION, NEURON_ID], but got {}.".format(
                    line
                )
            )
        edge_exists = action[0] != "!"

        return (u, v, self.ACTIONS[action[-1]], edge_exists)
